import networkx as nx
import pdb
import random
import communities as community
import pickle
import readdb as db
import sys

def positive(gr):
    # return a a graph with only positive edges
    ebunch = filter(lambda x:x[2]['sign'] < 0,gr.edges(data=True))
    gr.remove_edges_from(ebunch)

def nodesineach(assignments):
    counter = {}
    v = assignments.values()
    counter[-1] = v.count(-1)
    counter[1] = v.count(1)
    return counter


def within(gr,assignments):
    within = {}
    for i in gr.edges(data=True):
        # compute the value of the assignments for the ends of the edges and see to which cluster they belong to
        s = assignments[i[0]]+assignments[i[1]]
        if s==-2 or s==2:
            # belong to the same cluster
            try:
                within[s/2] +=1
            except KeyError:
                within[s/2] = 1
    # get the counter
    c = nodesineach(assignments)
    # divide by the square of the number
    for i in [-1,1]:
        within[i] = float(within[i])/c[i]**2
    return within,c

def links(gr,assignments,i):
    # get the neighbors of the node i
    nbrs = gr.neighbors(i)
    l = {1:0,-1:0}
    for i in nbrs:
        try:
            l[assignments[i]] +=1
        except KeyError:
            l[assignments[i]] =1
    return l

def distance(inter,csize,interlinks):
    d = {}
    for i in [-1,1]:
        d[i] = float(-2*interlinks[i])/csize[i] + inter[i]
    return d

def component(d):
    # return the component to which the node belongs to
    if d[-1] < d[1]:
        return -1
    else:
        return 1

def objective(gr,assignments):
    # compute the value of the frustation
    c1 = filter(lambda x:assignments[x]==-1,assignments.keys())
    c2 = set(gr.nodes()).difference(c1)
    com1 = dict(zip(c1,[-1]*len(c1)))
    com2 = dict(zip(c2,[1]*len(c2)))
    return community.compute_objective(gr,com1,com2)

def kernelkmeans(gr,iter=10):
    # returns two clusters that have a high density of positive edges
    # get the graph with only positive edges
    positive(gr)
    assignments = {} # -1 if the node belongs to first cluster and 1 if it belongs to second cluster
    # get initial clustering
    n1 = random.sample(gr.nodes(),len(gr.nodes())/2)
    c1 = dict(zip(n1,[1]*len(n1))) # cluster 1
    n2 = set(gr.nodes()).difference(n1)
    c2 = dict(zip(n2,[-1]*len(n2))) # cluster 2
    assignments = dict(c1.items()+c2.items())
    frus = {}
    f = objective(gr,assignments)
    frus[0] = f
    # start the iteration
    for it in range(1,iter):
        inter,csize = within(gr,assignments)
        nextassignments= {}
        for i in gr.nodes():
            # compute the 3rd term for each cluster
            # compute the number of links in each of the cluster
            interlinks = links(gr,assignments,i)
            # distance to mean in each cluster
            d = distance(inter,csize,interlinks)
            # find the cluster assignments
            c = component(d)
            nextassignments[i] = c
        # iteration is done
        assignments = nextassignments
        # find the value of the frustation
        f = objective(gr,assignments)
        frus[it] = f
        print it," ",f
        if f < 1000:
            break
    # write the value of the frustation in a file
    pickle.dump(frus,open("frustation_kmeans.obj","w"))
 
if __name__=='__main__':
    if len(sys.argv) < 2:
        print "Usage : python script input"
    else:
        gr = db.read_epinion(sys.argv[1])
        #################un comment the next line to work on the undirected part
        gr = db.undirected_graph(gr)
        kernelkmeans(gr,100)
