# computes the modularity of the given signed network in the form of a graph
# uses csr matrix format for fast matrix vector multiplication

import networkx as nx
import pdb
import sys
from scipy.sparse import *
from scipy import *
import math
import operator
import pickle
import visualization as visual
import triangles as triads
import sparse as S
import numpy as np
import random
import klin as lin
import frustree as optimum # optimum frustration value using post order traversal

np.seterr(invalid='raise')

def entry(x):
    return (x[0],x[1],x[2]['sign'])

def generate_sparse(gr):
    # gr is a networkx graph, return a csr matrix 
    neg_degree = [0]*len(gr)
    pos_degree = [0]*len(gr)
    data = [entry(x) for x in gr.edges(data=True)]
    rows = [x[0] for x in data]
    trows = list(rows)
    cols = [x[1] for x in data]
    entries = array([x[2] for x in data]*2)
    rows.extend(cols) # to make it symmetric matrix
    cols.extend(trows)
    ### generate the degree vectors also
    for i in gr.edges(data=True):
        try:
            if i[2]['sign'] == 1:
                pos_degree[i[0]] +=1
                pos_degree[i[1]] +=1
            else:
                neg_degree[i[0]] +=1
                neg_degree[i[1]] +=1
        except IndexError:
            print i
            pdb.set_trace()
            print "error"
            continue
    return array(pos_degree),array(neg_degree),csr_matrix( (entries,(rows,cols)),shape=(len(gr),len(gr)),dtype=int8)



def gen_pos_sparse(gr):
    ### generate B matrix
    pos_degree = [0]*len(gr)
    pos_edges = filter(lambda x:x[2]['sign']==1,gr.edges(data=True))
    data = [1]*(2*len(pos_edges))
    ## compute the rows and columns of the data
    rows = [i[0] for i in pos_edges]
    trows = list(rows)
    cols = [i[1] for i in pos_edges]
    ### make it symmetric
    rows.extend(cols) # to make it symmetric matrix
    cols.extend(trows)
    ### generate the degree vectors
    for i in pos_edges:
        pos_degree[i[0]]+=1
        pos_degree[i[1]]+=1
    return array(pos_degree),csr_matrix( (data,(rows,cols)),shape=(len(gr),len(gr)),dtype=int8)

def gen_neg_sparse(gr):
    ### generate B matrix
    neg_degree = [0]*len(gr)
    neg_edges = filter(lambda x:x[2]['sign']==-1,gr.edges(data=True))
    data = [1]*(2*len(neg_edges))
    ## compute the rows and columns of the data
    rows = [i[0] for i in neg_edges]
    trows = list(rows)
    cols = [i[1] for i in neg_edges]
    ### make it symmetric
    rows.extend(cols) # to make it symmetric matrix
    cols.extend(trows)
    ### generate the degree vectors
    for i in neg_edges:
        neg_degree[i[0]]+=1
        neg_degree[i[1]]+=1
    return array(neg_degree),csr_matrix( (data,(rows,cols)),shape=(len(gr),len(gr)),dtype=int8)

def magnitude(x):
    return sqrt(dot(x,x.conj()))

def normalize(vec):
    mag = magnitude(vec)
    #maxval = max(vec)
    #return maxval,vec/float(maxval)
    if mag==0:
        return 0.0,vec
    return mag,vec/float(mag)

def initial(n):
    # a matrix of size n*1 which stores the eigen vector
    #eigapprox = array([0]*n) # an eigen vector of size n
    #eigapprox[0] = 1
    #eigapprox = array([random.random() for i in range(n)])
    eigapprox = array([0.5 for i in range(n)])
    m,eigapprox = normalize(eigapprox)
    return m,eigapprox

def vec_mul(p,xvec):
    dot_val = dot(p.conj(),xvec)
    return p*dot_val

class Cluster:

    def __init__(self,pd,nd,adj,padj,nadj,vertices,pedges,nedges):
        self.pd = pd
        self.nd = nd
        self.padj = padj ### sparse matrix representation of only positive edges
        self.nadj = nadj ### sparse matrix representation of only positive edges
        self.adj = adj
        self.vertices = vertices
        self.pedges = pedges ## twice the number of edges in the graph..this is a multiplicative factor:
        self.nedges = nedges ## twice the number of edges in the graph..this is a multiplicative factor:


    def get_neg_frustration(self,gr):
        # compute the frustration contirbuted by the negative edges within this network
        H = gr.subgraph(self.vertices)
        fval = 0
        for i in H.edges(data=True):
            if i[2]['sign'] == -1:
                fval +=1
        return fval

    def myprint(self):
        print "Vertices are "
        if len(self.vertices) > 100:
            print self.vertices[:100]
            print "................."
        else:
            print self.vertices


    def is_not_trivial(self):
        ## return true if the number of nodes in the cluster is greater than 1
        return len(self.vertices)>1
    def multiply_vector(self,x):
        ## return adj+np-pp * x
        sp_prod = self.adj*x
        if self.nedges > 0:
            sp_prod = sp_prod+vec_mul(self.nd,x)/float(self.nedges)
        if self.pedges > 0:
            sp_prod = sp_prod-vec_mul(self.pd,x)/float(self.pedges)
        return sp_prod

    def bij_row_sum(self):
        ##
        pass

    def compute_psum(self,svector):
        ## compute the dot product for the klin algorithm
        if self.pedges > 0:
            return dot(self.pd.conj(),svector)/float(self.pedges)
        return 0
    def compute_nsum(self,svector):
        ## compute the dot product for the klin algorithm
        if self.nedges > 0:
            return dot(self.nd.conj(),svector)/float(self.nedges)
        return 0

    def sub_cluster(self,indices,vertices):
        ##
        arr_indices  = array(indices)
        ### new adjacency matrix
        newadj = S.pull(self.adj,arr_indices,arr_indices)
        newpadj = S.pull(self.padj,arr_indices,arr_indices)
        newnadj = S.pull(self.nadj,arr_indices,arr_indices)
        return Cluster(self.pd[arr_indices],self.nd[arr_indices],newadj,newpadj,newnadj,vertices,self.pedges,self.nedges)

    def modularity_dec(self,unit):
        ## decrease in the modularity
        ## this is the change compared to the two class case
        if self.pedges > 0:
            return self.padj*unit- ( vec_mul(self.pd,unit)/float(self.pedges) )
        return 0*unit

    def pos_diag(self):
        sz = len(self.vertices)
        unit = array([1]*sz)
        if self.pedges > 0:
            return self.padj*unit- ( vec_mul(self.pd,unit)/float(self.pedges) )
        return 0*unit
    def neg_diag(self):
        sz = len(self.vertices)
        unit = array([1]*sz)
        if self.nedges > 0:
            return self.nadj*unit- ( vec_mul(self.nd,unit)/float(self.nedges) )
        return 0*unit

    def get_vertices(self,eigapprox,factor=1):
        ## factor decides whether to return nodes belonging to cluster 1 or cluster -1
        if factor>0:
            verts = filter(lambda x:x[1]*factor >=0,enumerate(eigapprox))
        else:
            verts = filter(lambda x:x[1]*factor >0,enumerate(eigapprox))
        return [x[0] for x in verts],[self.vertices[x[0]] for x in verts]

    def get_cluster(self,ind,ver):
        ## get a cluster with ind and vertices
        if ind:
            ## if number of vertices > 0
            return self.sub_cluster(ind,ver)
    def compute_eig2(self,punit,nunit,shift,eigapprox):
        actual = self.multiply_vector(eigapprox)
        pfactor = punit*eigapprox
        nfactor = nunit*eigapprox
        eignext = actual-pfactor+nfactor+shift*eigapprox
        den = sum(eigapprox*eigapprox)
        if den !=0:
            return sum(eignext*eigapprox)/sum(eigapprox*eigapprox)
        else:
            return 0
        #return num/float(den)

    def compute_max_eigen(self,shift=0,ep=0.005):
        ## try to split a given cluster
        ## first find the maximum absolute valued eigen value
        ## if it its greater then 0 break
        ## else redo with a new matrix
        sz = len(self.vertices)
        m,eigapprox = initial(sz)
        m,eigapprox = normalize(eigapprox)
        prev_approx = None
        punit = self.pos_diag()
        nunit = self.neg_diag()
        iterations = 1
        while True:
            prev_approx = eigapprox
            ## update the eigen approx
            actual = self.multiply_vector(eigapprox)
            pfactor = punit*eigapprox
            nfactor = nunit*eigapprox
            eigapprox = actual-pfactor+nfactor+shift*eigapprox
            m,eigapprox = normalize(eigapprox)
            eigval = compute_eigenvalue(eigapprox,prev_approx)
            conv = convergence(eigapprox,eigval*prev_approx)
            if conv<=ep or iterations > 10000:
                ### compute the actual eigen value
                print "iterations ",iterations
                eigval = self.compute_eig2(punit,nunit,shift,eigapprox)
                return eigval,eigapprox
            iterations = iterations + 1

    def zeross(self,eigvec,ep=0.0):
        ## get the indices that are close to zero
        verts = filter(lambda x:0.0 < abs(x[1]) <=ep,enumerate(eigvec))
        return [x[0] for x in verts],[self.vertices[x[0]] for x in verts]

    def split_cluster(self,eigapprox):
        ## get two clusters by splitting the cluster using this eigen value
        pi,pv = self.get_vertices(eigapprox,1)
        ni,nv = self.get_vertices(eigapprox,-1)
        zi,zv = self.zeross(eigapprox,0.1)
        #print "eig approx",eigapprox
        #pdb.set_trace()
        ### now improve the clustering using the kerninghan
        #print "eig is ",eigapprox
        new_pi,new_pv,new_ni,new_nv = lin.improve(self,pi,ni,zi,pv,nv)
        #new_pi,new_pv,new_ni,new_nv = pi,pv,ni,nv
        #print "positive change ",len(set(pv).difference(new_pv))
        #print "negative change ",len(set(nv).symmetric_difference(new_nv))
        #c1,c2 = self.get_cluster(new_pi,new_pv),self.get_cluster(new_ni,new_nv)
        c1,c2 = self.get_cluster(pi,pv),self.get_cluster(ni,nv)
        #pdb.set_trace()
        #print "cluster sizes are ",len(self.vertices),len(c1.vertices),len(c2.vertices)
        #self.myprint()
        #print "divided into "
        #c1.myprint()
        #c2.myprint()
        return True,c1,c2

    def analyze_cluster(self):
        ## compute the eig value
        #print "analyzing ",self.vertices
        ep = math.pow(10,-6)
        eigval,eigapprox = self.compute_max_eigen(0,ep)
        #print " eigval and vector is ",eigval,eigapprox
        if eigval<0:
            ## compute it again using the modified matrix
            neig,napprox = self.compute_max_eigen(abs(eigval),ep)
            if neig-abs(eigval)>0:
                ## splitting is possible
                print neig-abs(eigval)
                return self.split_cluster(eigapprox)
            else:
                print "cannot be split any further "
                print "----------------------------"
                return False,[-1],[-1]
        else:
            ## divide the cluster into two parts and work on the individual
            ## pi -- positive indexand pv -- positive vertices llly negative
            return self.split_cluster(eigapprox)

    def frustration_value(self,gr):
        # compute the frustration of this cluster in the graph gr
        pass
    def process_cluster(self,iter=50,ep=.005):
        sz = len(self.vertices)
        ## see if this cluster can be divided further
        m,eigapprox = initial(sz)
        prev_approx  = -1
        prevmax  = 1
        ##### a unit vector for extra term
        punit = self.pos_diag()
        nunit = self.neg_diag()
        #pdb.set_trace()
        iterations = 0
        x = raw_input()
        inp = int(x)
        while True:
            prev_approx = eigapprox
            ## find the leading eigen vector and the pair
            ### compute the sum of Bij and Cij for each row
            actual = self.multiply_vector(eigapprox)
            pfactor = punit*eigapprox
            nfactor = nunit*eigapprox
            eigapprox = actual-pfactor+nfactor
            m,eigapprox = normalize(eigapprox)
            eigval = compute_eigenvalue(eigapprox,prev_approx)
            conv = convergence(eigapprox,eigval*prev_approx)
            print "------------------------------"
            print eigapprox
            print prev_approx
            print "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$"
            if conv<ep:
                break
        eigval = compute_eigenvalue(eigapprox,prev_approx)
        print "----------",eigval
        ## is the cluster divisible
        if eigval > 0:
            ## return two new sub clusters
            ## first find the vertices in both the clusters
            pos_indices = filter(lambda x:eigapprox[x]>=0,range(sz)) 
            pos_vertices = [self.vertices[x] for x in pos_indices]
            neg_indices = filter(lambda x:eigapprox[x]<0,range(sz)) 
            neg_vertices = [self.vertices[x] for x in neg_indices]
            c1 = None
            c2 = None
            if len(pos_indices)>0:
                c1 = self.sub_cluster(pos_indices,pos_vertices)
            if len(neg_indices)>0:
                c2 = self.sub_cluster(neg_indices,neg_vertices)
            return True,c1,c2
            
            ### create the sub matrices and the vectors
        else:
            return False,[-1],[-1]



def compute_eigenvalue(v1,v2):
    ## v1 is the latest value and the v2 is the previous value
    return v1.dot(v2)/float(v2.dot(v2))

def error(c2,c1):
    ## c2 is the current max
    print "------------------",c2,c1
    return abs( (c2-c1)/float(c2))*100

def convergence(eigapprox,prev_approx):
    ## compute the magnitude of the difference of the vectors
    d = eigapprox - prev_approx
    #print "errorrrrrrrrrr",math.sqrt(d.dot(d))
    return math.sqrt(d.dot(d))


def positive_between(c1,c2,gr):
    # return the number of positive egdes between the clusters c1 and c2 in the graph gr
    c1vertices = set(c1.vertices)
    c2vertices = set(c2.vertices)
    allvertices = c1vertices.union(c2.vertices)
    H = gr.subgraph(allvertices)
    fval = 0
    for i in H.edges(data=True):
        if i[2]['sign']==1 :
            ## check if the edge is coming from different communities
            val1  = 1 ; val2 = 1; # val1 is 1 if the source vertex is in cluster 1 and -1 o/w
            if i[0] in c2.vertices:
                val1 = -1
            if i[1] in c2.vertices:
                val2 = -1
            fval += 1*((1-val1*val2)/2)
    return fval
def multiple_clusters(gr,iter=10):
    pd,nd,aij_matrix = generate_sparse(gr) # generates the signed version of the adjacency matrix
    pd,padj_matrix   = gen_pos_sparse(gr)
    nd,nadj_matrix   = gen_neg_sparse(gr)
    #### maintain a list of clusters and process them until they cannot be divided any further
    clusters = [Cluster(pd,nd,aij_matrix,padj_matrix,nadj_matrix,gr.nodes(),sum(pd),sum(nd))] ### initial all the nodes belong to the same cluster
    final_clusters = []
    ### initialize the tree for the optimal clustering
    treemapping = {} # key is the cluster object and the value is the node in optimal tree
    initialNeg = clusters[0].get_neg_frustration(gr)
    tree = optimum.Tree(initialNeg)
    treemapping[clusters[0]] = tree
    root = tree
    while True and len(clusters) > 0:
        ## pop the first element and see if it can be divided
        c = clusters.pop(0)
        clustemp = c
        #pdb.set_trace()
        if not c:
            continue
        if not c.is_not_trivial():
            divisible = False
        else:
            divisible,c1,c2 = c.analyze_cluster()
        if divisible:
            ## append c1 and c2
            ## if both the clusters are non zero
            if (c1 and c2) and (len(c1.vertices) > 10 and len(c2.vertices) > 10):
                print "division"
                clusters.append(c1)
                clusters.append(c2)
                ## compute the positive value and also add the child and update the child nodes in the treemapping
                #l,r = treemapping[c].add_children(c1.get_neg_frustration(gr),c2.get_neg_frustration(gr))
                #treemapping[c1] = l; treemapping[c2] = r; ## add the nodes corresponding to the children in the treemap
                # compute the positve between and set the positive value for the parent 
                #pval = positive_between(c1,c2,gr)
                #treemapping[c].set_pval(pval)
                #pdb.set_trace()
            else:
                final_clusters.extend(filter(lambda x:x,[c1,c2]))
                #treemapping[c].set_pval(0) # as the cluster cannot be split we set the pval of the cluster to be zero
        else:
            final_clusters.append(c)
            #treemapping[c].set_pval(0) # as the cluster cannot be split we set the pval of the cluster to be zero
    return final_clusters,treemapping,tree.min_cut(),[i for i in tree.marked_nodes()],tree # return the clusters,treemapping, mincut and the tree

def delta(x,y):
    if x==y:
        return 1
    return 0
def frustation_val(clus,gr):
    ## compute the final frustation value in the graph
    print len(clus)
    #pdb.set_trace()
    fval = 0
    clumap = {}
    for index,item in enumerate(clus):
        d = dict(zip(item.vertices,[index]*len(item.vertices)))
        clumap = dict(clumap.items() + d.items())
    
    for i in gr.edges(data=True):
        if i[2]['sign'] == 1:
            fval += 1-delta(clumap[i[0]],clumap[i[1]])
        else:
            fval += delta(clumap[i[0]],clumap[i[1]])
    return fval


def iterative(gr,iter=10):
    # compute the eigen vector using the iterative method 
    fvals=[]
    m,eigapprox = initial(len(gr))
    pd,nd,aij_matrix = generate_sparse(gr) # generates the signed version of the adjacency matrix
    iterations = 0
    print "Computed the sparse matrix"
    oldval = frustated_value(gr,sorted_indices(eigapprox))
    fvals.append(oldval)
    while True and iterations < iter:
        print "Iteration no :",iterations
        sp_prod = aij_matrix*eigapprox
        np_val = vec_mul(nd,eigapprox)
        pp_val = vec_mul(pd,eigapprox)
        eigapprox = sp_prod + np_val - pp_val
        m,eigapprox = normalize(eigapprox)
        iterations +=1
        #newval = frustated_value(gr,sorted_indices(eigapprox))
        newval = compute_frustation(gr,eigapprox)
        if newval==oldval:
            break
        else:
            oldval= newval
            fvals.append(newval)
        print newval
    print "got the top eigen vector"
    return fvals,eigapprox

def sorted_indices(v): 
    return [ i for (i,j) in sorted(enumerate(v), key=operator.itemgetter(1))]

def assignments(v):
    ## given a eigen vector assign all the nodes with -ve score to cluster 2 and positive ones to c1
    pindices = filter(lambda x:v[x] >=0, range(len(v)))
    nindices = filter(lambda x:v[x] < 0, range(len(v)))
    c1 = dict(zip(pindices,[1]*len(pindices)))
    c2 = dict(zip(nindices,[-1]*len(nindices)))
    merge_dict = dict(c1.items()+c2.items())
    return merge_dict

def compute_frustation(gr,v):
    ## v is an eigen vector
    ## first compute the assignments and then compute the frustation
    merge_dict = assignments(v)
    pedges = filter(lambda x:x[2]['sign']==1,gr.edges(data=True))
    nedges = filter(lambda x:x[2]['sign']==-1,gr.edges(data=True))
    pcount = len(filter(lambda x:merge_dict[x[0]]*merge_dict[x[1]]==1,pedges))
    ncount = len(filter(lambda x:merge_dict[x[0]]*merge_dict[x[1]]==-1,nedges))
    return pcount+ncount,len(gr.edges())-(pcount+ncount)

def frustated_value(gr,v):
    fval = 0
    nval = 0
    c1 = v[:len(v)/2]
    c2 = v[len(v)/2:]
    c1 = dict(zip(c1,[1]*len(c1)))
    c2 = dict(zip(c2,[-1]*len(c2)))
    merge_dict = dict(c1.items()+c2.items())
    pedges = filter(lambda x:x[2]['sign']==1,gr.edges(data=True))
    nedges = filter(lambda x:x[2]['sign']==-1,gr.edges(data=True))
    pcount = len(filter(lambda x:merge_dict[x[0]]*merge_dict[x[1]]==1,pedges))
    ncount = len(filter(lambda x:merge_dict[x[0]]*merge_dict[x[1]]==-1,nedges))
    return pcount+ncount,len(gr.edges())-(pcount+ncount)

def fval_change(frustu):
    visual.plotpoints( frustu.keys(),frustu.values())
    visual.saveplot('communities.png',"Change in the frustation value","Iteration #","Frustation value")


def triads_modularity(gr,merge_dict,triadType):
    ## for a given traid type - - - or + + - see how many of frustation edges participate in the a given type
    tri = triads.get_triads(gr,triadType)
    ## for each of this triad see how many of these have two vertices in one cluster and the third in other cluster
    multiple_valid = 0
    total = 0
    for i in tri:
        ## if triadType == 0  then the sum cluster assignments is 1 or -1
        s = abs(sum(map(lambda x:merge_dict[x],i)))
        if s==1:
            multiple_valid +=1
        total +=1
    return multiple_valid/float(total),total


def runner(g,iter=10):
    # run the entire algorithm get the clusters and the frustation and the negative value and return
    fvals,topeig = iterative(g,iter)
    #clusters = sorted_indices(topeig)
    #f,n = frustated_value(g,clusters)
    f,n = compute_frustation(g,topeig)
    ###### check if the multiple clusters is a valid approach for this
    frac,tot = triads_modularity(g,assignments(topeig),0)
    pickle.dump([fvals,topeig,f,n,(frac,tot)],open('modularity.obj','w'))
    frustvals = dict(zip(range(len(fvals)),fvals))
    fval_change(frustvals)
    return f,n,topeig
    
if __name__=='__main__':
    g = nx.Graph()
    g.add_edge(0,1,sign=-1)
    g.add_edge(1,2,sign=1)
    g.add_edge(0,2,sign=-1)
    #pd,nd,sp = generate_sparse(g)
    #x = initial(3)
    topeig = iterative(g)
    clusters = sorted_indices(topeig)
    f,n = frustated_value(g,clusters)
    #pdb.set_trace()
