# spectral algorithm to minimize frustration value
import networkx as nx
import pdb
import sys
from scipy.sparse import *
from scipy import *
import math
import sparse as sp


def assign_vector(eigapprox):
    # return the assignment vector based on leading eigen vector
    ar = array(eigapprox)
    return transpose(matrix(map(round,ar))) # round each element and return a column vector

def round(x):
    if x>0:
        return 1
    return -1

def initial(n):
    # a matrix of size n*1 which stores the eigen vector
    eigapprox = array([0.5 for i in range(n)])
    return transpose(matrix(eigapprox))

def normalize(eigapprox):
    # compute the max magnitude element in this column vector
    ar = array(eigapprox)
    mval = max(abs(ar))
    if mval==0:
        #pdb.set_trace()
        return [None,None]
    return mval,eigapprox/float(mval)

def gen_sparse(gr):
    # generate a signed sparse matrix of the graph
    #return nx.to_scipy_sparse_matrix(gr,weight='sign',format = 'csr')
    # in the order returned by gr.nodes()
    return nx.to_scipy_sparse_matrix(gr).tocsr()
class Fcluster:

    def __init__(self,adjmatrix,verts):
        
        # adjmatrix is the sparse matrix representation of the grpah induced by this subgraph
        # csr matrix format is faster for matrix vector multiplications
        self.adjmatrix = adjmatrix 
        self.vertices = verts
        self.length = len(verts)

    def compute_eigen(self):
        # compute the leading eigen vector of the signed matrix represented by this set of vertices
        ep = math.pow(10,-4)
        eigapprox = initial(self.length)
        prev_val,eigapprox = normalize(self.adjmatrix * eigapprox)
        ## if prev_val is None then retun Null
        if not prev_val:
            return [None,None]
        iterations = 0
        print "Number of vertices ",self.length
        while True:
            curr_val,eigapprox = normalize(self.adjmatrix * eigapprox)
            ## if the max abs value is 0..it returns None
            if not curr_val:
                return [None,None]
            # check for convergenve
            if abs ( (curr_val-prev_val)/float(curr_val)) <= ep or iterations>2000:
                #print curr_val
                return curr_val,eigapprox
            # update the previous value
            prev_val = curr_val
            iterations = iterations + 1
            #print iterations
            


    def maxvalue(self,assign):
        # return the max value of S.t * A * S where A is the signed adjacency and S is the assignment vector
        # assign is column vector
        f1 = self.adjmatrix * assign
        res = transpose(assign) * f1
        return res[0,0]

    def gen_cluster(self,indices):
        # generate a cluster with these indices
        return Fcluster(sp.pull(self.adjmatrix,array(indices),array(indices)),[self.vertices[x] for x in indices])

    def split_cluster(self,assignment):
        ## return two new clusters based on the assignment vector
        pindex = filter(lambda x:assignment[x] == 1,range(len(assignment)))
        nindex = filter(lambda x:assignment[x] == -1,range(len(assignment)))

        if len(pindex) == 0 or len(nindex) == 0:
            return False,self
        else:
            # generate clusters from positive and negative clusters
            pclus = self.gen_cluster(pindex)
            nclus = self.gen_cluster(nindex)
            return True,pclus,nclus

        
    def ambiguous_vertices(self,eigapprox,threshold=0.1):
        # return the indices of vertices on which improvement has to be applied
        ar = array(eigapprox)
        return [k for k,v in enumerate(ar) if 0.0 < abs(v) < threshold]

    def test_split(self):

        # change in the frustration value by spltting this cluster is given by delta f = sigma_{ij} aij*(0.5-0.5*s_{i}*s_{j})
        # 1. compute the leading eigen vector 
        # 2. assign the nodes based on the eigen vector
        # 3. if the change is negtive then split this cluster into two clusters
        
        #### compute the sum of elements in the signed adjacency matrix
        constval = sum(self.adjmatrix.data)
        #### first compute the leading eigen vector of the adjacency matrix
        eigval,eigapprox = self.compute_eigen()
        ### if eigen cannot be computed...then return value is [None,None]
        if not eigval:
            return [False]
        print "eigen value is ",eigval
        ### get the ambiguous vertices for this eigenvector
        av = self.ambiguous_vertices(eigapprox)
        #### get the assign column matrix from the eigapprox
        assignment = assign_vector(eigapprox)
        #### compute the product S.t * A * S
        mval = self.maxvalue(assignment)
        #### check if the overall frustration value decreases
        if constval > mval :
            return [False]
        else:
            # if the value decreases then split this cluster based on the assignment vector
            ret = self.split_cluster(assignment)
            if ret[0]:
                return True,ret[1],ret[2],av
            else:
                return [False]


def optimal_frustration(gr):
    queue = [Fcluster(gen_sparse(gr),gr.nodes())]
    final_clusters = []
    while queue:
        c = queue.pop(0)
        if len(c.vertices) == 1:
            final_clusters.append(c)
            continue
        ret = c.test_split()
        if ret[0]:
            # this cluster cannot be split further
            print "division ",len(queue)
            #### stop the process at two clusters
            ret = improve(c,ret[1],ret[2],ret[3],gr)
            if ret[0]:
                final_clusters.extend([ret[1],ret[2]])
            # ret[3] is the list of indices on which improvement has to be performed
            #ret = improve(c,ret[1],ret[2],ret[3],gr)
            #if ret[0] : # if the improvement doesn't lead to a degenerate case..i.e placing all the vertices in a single cluster
            #   queue.extend([ret[1],ret[2]])
            #if ret[0] : # if the improvement doesn't lead to a degenerate case..i.e placing all the vertices in a single cluster
            #  queue.extend([ret[1],ret[2]])
            else:
                final_cluster.append(c)
        else:
            final_clusters.append(c)
    ## compute the frustration with these clusters
    #return compute_frustration(gr,final_clusters)
    fval= compute_frustration(final_clusters,gr)
    return final_clusters,fval

def min_key(a):
    ## find the key for which the value is the maximum
    b = dict(map(lambda item: (item[1],item[0]),a.items()))
    return b[min(b.keys())]
def improve(c,c1,c2,av,gr):
    print "Improvement with ",len(av)
    # move the vertices so as to minimize the frustrations
    c1map = dict(zip(c1.vertices,[1]*c1.length))
    c2map = dict(zip(c2.vertices,[-1]*c2.length)) # vertices that belong to cluster 2
    cmap = dict(c1map.items()+c2map.items())
    #print "Initial ",cmap
    cmap_bkp = cmap.copy()
    #unmoved = set(c1.vertices).union(c2.vertices)
    # perform the improvement only on ambiguous vertices
    unmoved = [c.vertices[x] for x in  av]
    improve = []
    indices = []
    allvertices = set(c.vertices)
    for i in range(len(unmoved)):
        score = {}
        for index,item in enumerate(unmoved):
            # compute the change by moving this vertex from its current to the new cluster
            # get the neighbors of item
            # consider only those neighbors that are present in  either cluster
            nbrs = set(gr.neighbors(item))
            nbrs.intersection_update(allvertices)
            srcC = cmap[item]
            change = 0
            for nbr in nbrs:
                if gr[item][nbr]['weight'] == 1: # if the edge is positive
                    if srcC * cmap[nbr] == 1 :  # both belong to the same cluster
                        change += 1 # both in same cluster to different clusters
                    else:
                        change += -1 # different clusters to same cluster
                else:
                    if srcC * cmap[nbr] == 1 :  # both belong to the same cluster
                        change += -1
                    else:
                        change += 1 # different clusters to same clusters
            score[item] = change
            if index > 5000:
                break
        # compute the least index
        minindex = min_key(score)
        cmap[minindex] = -1*cmap[minindex]
        indices.append(minindex)
        try:
            improve.append(improve[-1]+score[minindex])
        except IndexError:
            improve.append(score[minindex])
        ### update the svector
        unmoved.remove(minindex)
        if i>5000:
            break
    # find the index where there is greatest change
    #m = min(improve)..consider the case where improve can be an empty list
    m = 1 if len(improve)==0 else min(improve)
    if m >= 0:
        return True,c1,c2
    else:
        # find the max decrease
        lastindex = (len(improve) - 1) - improve[::-1].index(m)
        for chng in indices[:lastindex+1]:
            cmap_bkp[chng] *=-1
        #pdb.set_trace()
        ## find the new clusters..first compute the new assignment vector
        return c.split_cluster([cmap_bkp[x] for x in c.vertices])
        
def delta(x,y):
    return 1 if x==y else 0
def compute_frustration(clusters,gr):
    # compute the total frustration of this splitting
    ### generate a dict with vertex as the key and cluster id as the value
    #vmapping = dict(reduce(lambda x,y:x.extend(y) or x,[dict(zip(vertices,[index]*len(vertices))).items() for index,vertices in enumerate([i.vertices for i in clusters])]))
    #vmapping = reduce(lambda x,y: x.update(y) or x,[dict(vertices,[index]*len(vertices) for index,vertices in enumerate([i.vertices for i in clusters])).items()])
    ### this version is efficient compared to the previous line... nested for in comprehensions
    vmapping = dict((vert, i) for (i, cl) in enumerate(clusters) for vert in cl.vertices)
    #print vmapping
    fval = 0
    for i in gr.edges(data=True):
        if i[2]['weight'] == 1:
            fval += 1-delta(vmapping[i[0]],vmapping[i[1]])
        else:
            fval += delta(vmapping[i[0]],vmapping[i[1]]) ## add 1 if the negative edge lies in the same cluster
    return fval

def pos_neg_deg(gr):
    # return a dict with node as the key and positive degree or negative degree as the value
    pos_deg = dict(zip(gr.nodes(),[0]*len(gr)))
    neg_deg = dict(zip(gr.nodes(),[0]*len(gr)))
    for i in gr.edges(data=True):
        if i[2]['weight'] == 1:
            pos_deg[i[0]] +=1;pos_deg[i[1]]+=1;
        else:
            neg_deg[i[0]] +=1;neg_deg[i[1]]+=1;
    return pos_deg,neg_deg

def compute_modularity(clusters,gr):
    ## get the vertex mapping
    vmapping = dict((vert, i) for (i, cl) in enumerate(clusters) for vert in cl.vertices)
    ## get the positive degree map and negative degree map from the graph
    pos_deg,neg_deg = pos_neg_deg(gr)
    Q = 0.0
    ## compute the positive and negative edges
    numpos = sum(pos_deg.values());numneg = sum(neg_deg.values());
    # get the total number of positive edges and negative edges
    #numpos = len(filter(lambda x:x[2]['weight']==1,gr.edges(data=True)))
    #numneg = gr.number_of_edges()-numpos
    pos_norm = 0; neg_norm = 0;
    if numpos > 0:
        pos_norm = 1.0/numpos
    if numneg > 0:     
        neg_norm = 1.0/numneg
    for u,v,d in gr.edges(data=True):
        # if the edge is positive and is between nodes in the same cluster
        edge_type = 1 if vmapping[u] == vmapping[v] else -1
        prods = pos_deg[u]*pos_deg[v]*pos_norm if edge_type==1 else neg_deg[u]*neg_deg[v]*neg_norm
        if d['weight']*edge_type == 1: # i.e positive egdes withing and negative edges across
            Q += 1 - prods
    pdb.set_trace()
    return Q


if __name__ =='__main__':
    print "---------"
    a = [1.5,0,1,-0.5,0.5,-0.5,-0.5,0,0]
    adj = matrix(a)
    adj.shape = 3,3
    clus = Fcluster(adj,[1,2,3])
    e,v = clus.compute_eigen()
    #print e
    gr = nx.Graph()
    gr.add_edge(0,1,weight=-1)
    gr.add_edge(1,2,weight=-1)
    gr.add_edge(2,0,weight=1)
    #sp = gen_sparse(gr)
    clus,fval = optimal_frustration(gr)
    pdb.set_trace()
